import torch.nn as nn
from lime import lime_image
from lime.wrappers.scikit_image import SegmentationAlgorithm


class Lime(object):
    '''

    '''
    def __init__(self,
                 model: nn.Module):
        self.model = model
        self.explainer = lime_image.LimeImageExplainer(verbose = False)
        self.segmenter = SegmentationAlgorithm('quickshift', kernel_size=1, max_dist=200, ratio=0.2)

    def __call__(self, x):
        explanation = self.explainer.explain_instance(x.to('cpu').numpy(),
                                                      self.model,
                                                      top_labels=5,
                                                      hide_color=0,
                                                      num_samples=1)        

class NN(nn.Module):
    def __init__(self):
        super(NN, self).__init__()
        self.net = torch.nn.Linear(32*32*3, 10)
        
    def forward(self, x):
        if isinstance(type(x), type(np.ndarray)):
            x = torch.Tensor(x)
        print(x.shape)
        x = x.view(x.size(0), -1)
        return self.net(x)
        
if __name__ == '__main__':
    import torch
    import numpy as np
    x = torch.rand(1, 32, 32, 3)
    net = NN()
    net(x)
    lime = Lime(net)
    lime(x[0][0])



import numbers

import torch
import torch.nn as nn


def chrono_init(rnn: torch.nn.Module, Tmax=None, Tmin=1):
    '''chrono initialization(Ref: https://arxiv.org/abs/1804.11188)
    '''
    
    assert isinstance(Tmin, numbers.Number), 'Tmin must be numeric.'
    assert isinstance(Tmax, numbers.Number), 'Tmax must be numeric.'    
    for name, p in rnn.named_parameters():
        if 'bias' in name:
            n = p.nelement()
            hidden_size = n // 4            
            p.data.fill_(0)
            if isinstance(rnn, (torch.nn.LSTM, torch.nn.LSTMCell)):
                p.data[hidden_size: 2*hidden_size] = \
                    torch.log(torch.nn.init.uniform_(p.data[0: hidden_size], 1, Tmax - 1))
                # forget gate biases = log(uniform(1, Tmax-1))
                p.data[0: hidden_size] = -p.data[hidden_size: 2*hidden_size]
                # input gate biases = -(forget gate biases)
    return rnn


def feedforward_init(dnn: nn.Module,
                     init_mean=0,
                     init_std=1,
                     init_xavier: bool=True,
                     init_normal: bool=True,
                     init_gain: float=1.0):
    for name, p in dnn.named_parameters():
        if 'bias' in name:
            p.data.zero_()
        if 'weight' in name:            
            if init_xavier:
                if init_normal:
                    nn.init.xavier_normal(p.data, init_gain)
                else:
                    nn.init.xavier_uniform(p.data, init_gain)
            else:
                if init_normal:
                    nn.init.normal(p.data, init_gain)
                else:
                    nn.init.uniform(p.data, init_gain)

                    
def rnn_init(rnn: nn.Module,
             init_xavier: bool=True,
             init_normal: bool=True,
             init_gain: float=1.0,
             init_mean: float=0.0,
             init_std: float=0.1,
             init_lower: float=0.0,
             init_upper: float=0.04):
    
    for name, p in rnn.named_parameters():
        if 'bias' in name:
            p.data.fill_(0)
            if isinstance(rnn, (torch.nn.LSTM, torch.nn.LSTMCell)):
                n = p.nelement()
                p.data[n // 4:n // 2].fill_(1)  # forget bias
        elif 'weight' in name:
            if init_xavier:
                if init_normal:
                    nn.init.xavier_normal(p, init_gain)
                else:
                    nn.init.xavier_uniform(p, init_gain)
            else:
                if init_normal:
                    try:
                        # from pytorch 4.0
                        nn.init.normal_(p, init_mean, init_std)
                    except:
                        # pytorch 3.1
                        nn.init.normal(p, init_mean, init_std)                        
                else:
                    try:
                        nn.init.uniform_(p, init_lower, init_upper)
                    except:
                        nn.init.uniform(p, init_lower, init_upper)                        
                    



# TODO: Conv2dLocal
# TODO: Conv2dMap
# TODO: ConvTranspose2dMap
import torch    
import torch.nn as nn
import torch.nn.functional as F
    
class MLPConv2d(nn.Module):
    """__init__(self, in_channels, out_channels, ksize=None, stride=1, pad=0, activation=relu.relu, conv_init=None, bias_init=None)
    Two-dimensional MLP convolution layer of Network in Network.
    This is an "mlpconv" layer from the Network in Network paper. This layer
    is a two-dimensional convolution layer followed by 1x1 convolution layers
    and interleaved activation functions.
    Note that it does not apply the activation function to the output of the
    last 1x1 convolution layer.
    Args:
        in_channels (int or None): Number of channels of input arrays.
            If it is ``None`` or omitted, parameter initialization will be
            deferred until the first forward data pass at which time the size
            will be determined.
        out_channels (tuple of ints): Tuple of number of channels. The i-th
            integer indicates the number of filters of the i-th convolution.
        ksize (int or pair of ints): Size of filters (a.k.a. kernels) of the
            first convolution layer. ``ksize=k`` and ``ksize=(k, k)`` are
            equivalent.
        stride (int or pair of ints): Stride of filter applications at the
            first convolution layer. ``stride=s`` and ``stride=(s, s)`` are
            equivalent.
        pad (int or pair of ints): Spatial padding width for input arrays at
            the first convolution layer. ``pad=p`` and ``pad=(p, p)`` are
            equivalent.
        activation (callable):
            Activation function for internal hidden units.
            You can specify one of activation functions from
            :doc:`built-in activation functions </reference/functions>` or
            your own function.
            It should not be an
            :doc:`activation functions with parameters </reference/links>`
            (i.e., :class:`~chainer.Link` instance).
            The function must accept one argument (the output from each child
            link), and return a value.
            Returned value must be a Variable derived from the input Variable
            to perform backpropagation on the variable.
            Note that this function is not applied to the output of this link.
            a keyword argument.
    .. note:
        From v2, `conv_init` and `bias_init` arguments must be specified as
        keyword arguments only. We impose this restriction to forbid
        users to assume the API for v1 and specify `wscale` option,
        that had been between `activation` and `conv_init` arguments in v1.
    See: `Network in Network <https://arxiv.org/abs/1312.4400v3>`_.
    Attributes:
        activation (callable):
            Activation function.
            See the description in the arguments for details.
    """  # NOQA

    def __init__(self, in_channels, out_channels, kernel_size=None, stride=1, padding=0,
                 activation=F.relu, *args, **kwargs):
        super(MLPConv2d, self).__init__()
        #if kernel_size is None:
        #    out_channels, kernel_size, in_channels = in_channels, out_channels, None

        assert len(out_channels) > 0
        
        self.convs = [nn.Conv2d(in_channels, out_channels[0], kernel_size, stride, padding)]
        for n_in, n_out in zip(out_channels, out_channels[1:]):
            self.convs.append(nn.Conv2d(n_in, n_out, 1))
        self.convs = nn.ModuleList(self.convs)
        self.activation = activation

    def forward(self, x):
        """Computes the output of the mlpconv layer.
        Args:
            x (~chainer.Variable): Input image.
        Returns:
            ~chainer.Variable: Output of the mlpconv layer.
        """
        f = self.activation
        for l in self.convs:
            x = f(l(x))
            
        return x
    
    
class UpsampleConvLayer(nn.Module):
    """UpsampleConvLayer
    Upsamples the input and then does a convolution. This method gives better results
    compared to ConvTranspose2d.
    ref: http://distill.pub/2016/deconv-checkerboard/
    """

    def __init__(self, in_channels, out_channels, kernel_size=1, stride=1, upsample=None):
        super(UpsampleConvLayer, self).__init__()
        self.upsample = upsample
        reflection_padding = kernel_size // 2
        self.reflection_pad = torch.nn.ReflectionPad2d(reflection_padding)
        self.conv2d = torch.nn.Conv2d(in_channels, out_channels, kernel_size, stride)

    def forward(self, x):
        x_in = x
        if self.upsample:
            x_in = torch.nn.functional.interpolate(x_in, mode='nearest', scale_factor=self.upsample)
        out = self.reflection_pad(x_in)
        out = self.conv2d(out)
        return out

    



import torch
import torch.nn as nn
import torch.nn.functional as F

from .lazy import Conv2d

class Conv2dBlock(nn.Module):
    def __init__(self, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1, bias=True,
                 dropout:float=0.0, pooling=False, batchnorm=False,
                 activation='relu'):
        super(Conv2dBlock, self).__init__()

        self.dropout = dropout
        self.pooling = pooling        
        self.batchnorm = batchnorm
        
        self.conv = Conv2d(out_channels, kernel_size, stride,
                           padding, dilation, transposed, output_padding, groups, bias),


    def forward(self, x):
        return self.model(x)
        
    


class Inception(nn.Module):
    """Inception module of GoogLeNet.
    It applies four different functions to the input array and concatenates
    their outputs along the channel dimension. Three of them are 2D
    convolutions of sizes 1x1, 3x3 and 5x5. Convolution paths of 3x3 and 5x5
    sizes have 1x1 convolutions (called projections) ahead of them. The other
    path consists of 1x1 convolution (projection) and 3x3 max pooling.
    The output array has the same spatial size as the input. In order to
    satisfy this, Inception module uses appropriate padding for each
    convolution and pooling.
    See: `Going Deeper with Convolutions <https://arxiv.org/abs/1409.4842>`_.
    Args:
        in_channels (int or None): Number of channels of input arrays.
        out1 (int): Output size of 1x1 convolution path.
        proj3 (int): Projection size of 3x3 convolution path.
        out3 (int): Output size of 3x3 convolution path.
        proj5 (int): Projection size of 5x5 convolution path.
        out5 (int): Output size of 5x5 convolution path.
        proj_pool (int): Projection size of max pooling path.

    TODO: Implement dimension reduction version.
    """

    def __init__(self, in_channels, out1, proj3, out3, proj5, out5, proj_pool):
        super(Inception, self).__init__()
        self.conv1 = nn.Conv2d(in_channels, out1, 1)
        self.proj3 = nn.Conv2d(in_channels, proj3, 1)
        self.conv3 = nn.Conv2d(proj3, out3, 3, padding=1)
        self.proj5 = nn.Conv2d(in_channels, proj5, 1)
        self.conv5 = nn.Conv2d(proj5, out5, 5, padding=2)
        self.projp = nn.Conv2d(in_channels, proj_pool, 1)

    def forward(self, x):
        """Computes the output of the Inception module.
        Args:
            x (~torch.Tensor): Input variable.
        Returns:
            Variable: Output variable. Its array has the same spatial size and
            the same minibatch size as the input array. The channel dimension
            has size ``out1 + out3 + out5 + proj_pool``.
        """
        out1 = self.conv1(x)
        out3 = self.conv3(F.relu(self.proj3(x)))
        out5 = self.conv5(F.relu(self.proj5(x)))
        pool = self.projp(F.max_pool2d(x, 3, stride=1, padding=1))
        y = F.relu(torch.cat((out1, out3, out5, pool), dim=1))
        return y


class InceptionBN(nn.Module):

    """Inception module of the new GoogLeNet with BatchNormalization.
    This chain acts like :class:`Inception`, while InceptionBN uses the
    :class:`BatchNormalization` on top of each convolution, the 5x5 convolution
    path is replaced by two consecutive 3x3 convolution applications, and the
    pooling method is configurable.
    See: `Batch Normalization: Accelerating Deep Network Training by Reducing \
    Internal Covariate Shift <https://arxiv.org/abs/1502.03167>`_.
    Args:
        in_channels (int or None): Number of channels of input arrays.
        out1 (int): Output size of the 1x1 convolution path.
        proj3 (int): Projection size of the single 3x3 convolution path.
        out3 (int): Output size of the single 3x3 convolution path.
        proj33 (int): Projection size of the double 3x3 convolutions path.
        out33 (int): Output size of the double 3x3 convolutions path.
        pooltype (str): Pooling type. It must be either ``'max'`` or ``'avg'``.
        proj_pool (int or None): Projection size in the pooling path. If
            ``None``, no projection is done.
        stride (int): Stride parameter of the last convolution of each path.
    .. seealso:: :class:`Inception`
    """

    def __init__(self, in_channels, out1, proj3, out3, proj33, out33,
                 pooltype='max', proj_pool=None, stride=1):
        super(InceptionBN, self).__init__()
        self.out1 = out1
        self.proj_pool = proj_pool
        self.stride = stride
        self.pooltype = pooltype
        if pooltype != 'max' and pooltype != 'avg':
            raise NotImplementedError()
        
        self.proj3 = nn.Conv2d(in_channels, proj3, 1, bias=False)
        self.conv3 = nn.Conv2d(proj3, out3, 3, padding=1, stride=stride, bias=False)
        self.proj33 = nn.Conv2d(in_channels, proj33, 1, bias=False)
        self.conv33a = nn.Conv2d(proj33, out33, 3, padding=1, bias=False)
        self.conv33b = nn.Conv2d(out33, out33, 3, padding=1, stride=stride, bias=False)
        self.proj3n = nn.BatchNorm2d(proj3)
        self.conv3n = nn.BatchNorm2d(out3)
        self.proj33n = nn.BatchNorm2d(proj33)
        self.conv33an = nn.BatchNorm2d(out33)
        self.conv33bn = nn.BatchNorm2d(out33)

        if out1 > 0:
            assert stride == 1
            assert proj_pool is not None
            self.conv1 = nn.Conv2d(in_channels, out1, 1, stride=stride, bias=False)
            self.conv1n = nn.BatchNorm2d(out1)

        if proj_pool is not None:
            self.poolp = nn.Conv2d(in_channels, proj_pool, 1, bias=False)
            self.poolpn = nn.BatchNorm2d(proj_pool)

    def forward(self, x):
        outs = []

        if self.out1 > 0:
            h1 = self.conv1(x)
            h1 = self.conv1n(h1)
            h1 = F.relu(h1)
            outs.append(h1)

        h3 = F.relu(self.proj3n(self.proj3(x)))
        h3 = F.relu(self.conv3n(self.conv3(h3)))
        outs.append(h3)

        h33 = F.relu(self.proj33n(self.proj33(x)))
        h33 = F.relu(self.conv33an(self.conv33a(h33)))
        h33 = F.relu(self.conv33bn(self.conv33b(h33)))
        outs.append(h33)

        if self.pooltype == 'max':
            p = F.max_pool2d(x, 3, stride=self.stride, padding=1,
                             cover_all=False)
        else:
            p = F.avg_pool2d(x, 3, stride=self.stride,
                             padding=1)
        if self.proj_pool is not None:
            p = F.relu(self.poolpn(self.poolp(p)))
        outs.append(p)

        y = torch.cat(outs, dim=1)
        return y
    



import math

import torch
import torch.nn as nn
from torch.nn import init
import torch.nn.functional as F

from .utils import _single, _pair, _triple
from .base import LazyBase

class _ConvNd(LazyBase):
    def __init__(self, out_channels, kernel_size, stride,
                 padding, dilation, transposed, output_padding, groups, bias):
        super(_ConvNd, self).__init__()
        if out_channels % groups != 0:
            raise ValueError('out_channels must be divisible by groups')
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.transposed = transposed
        self.output_padding = output_padding
        self.groups = groups
        
        self.in_channels = None
        self.weight = nn.Parameter(None)
        self.bias = nn.Parameter(None)
        self.bias_flag = bias
        self._register_load_state_dict_pre_hook(self._lazy_load_state_dict_hook)        

    def _initialize_weight(self, in_channels):
        self.in_channels = in_channels
        if self.in_channels % self.groups != 0:
            raise ValueError('in_channels must be divisible by groups')

        if self.transposed:
            self.weight.data = torch.Tensor(self.in_channels,
                                            self.out_channels // self.groups,
                                            *self.kernel_size)
        else:
            self.weight.data = torch.Tensor(self.out_channels,
                                            self.in_channels // self.groups,
                                            *self.kernel_size)
                
        if self.bias_flag:
            self.bias.data = torch.Tensor(self.out_channels)

        self.weight = self._to_device(self.weight)
        self.bias = self._to_device(self.bias)            
        self._reset_parameters()

    def _reset_parameters(self):
        n = self.in_channels
        init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
            bound = 1 / math.sqrt(fan_in)
            init.uniform_(self.bias, -bound, bound)

    def _lazy_load_state_dict_hook(self, state_dict, prefix, local_metadata, strict,
                                   missing_keys, unexpected_keys, error_msgs):
        for name, data in state_dict.items():
            if prefix in name:
                if 'weight' in name:
                    self.in_channels = data.shape[0]
                    self.weight.data = data
                elif 'bias' in name:
                    self.bias.data = data
                else:
                    raise ValueError(name)
            
    def extra_repr(self):
        s = ('{in_channels}, {out_channels}, kernel_size={kernel_size}'
             ', stride={stride}')
        if self.padding != (0,) * len(self.padding):
            s += ', padding={padding}'
        if self.dilation != (1,) * len(self.dilation):
            s += ', dilation={dilation}'
        if self.output_padding != (0,) * len(self.output_padding):
            s += ', output_padding={output_padding}'
        if self.groups != 1:
            s += ', groups={groups}'
        if self.bias is None:
            s += ', bias=False'
        return s.format(**self.__dict__)


class Conv1d(_ConvNd):
    '''
    :param out_channels: the size of the window to take a max and average over
    :param kernel_size: the size of the window to take a max and average over
    :param stride: the size of stride to move kernel
    :param padding: implicit zero padding to be added on both sides
    :param dilation: a parameter that controls the stride of elements in the window
    :param return_indices: if True, will return the max indices along with the outputs. Useful when Unpooling later
    :param ceil_mode: when True, will use ceil instead of floor to compute the output shape    

    :type kernel_size: int or list
    :type stride: int or list
    
    Examples::

        import torch
        import torchex.nn as exnn
     
        net = exnn.Conv1d(10, 2)
        # You don't need to give the size of input for this module.
        # This network is equivalent to `nn.Conv1d(3, 10, 2)`.
     
        x = troch.randn(10, 3, 28)
        y = net(x)
    '''
    def __init__(self, out_channels: int, kernel_size: int or list, stride: int or list=1,
                 padding: int=0, dilation: int=1, groups: int=1, bias: bool=True):
        kernel_size = _single(kernel_size)
        stride = _single(stride)
        padding = _single(padding)
        dilation = _single(dilation)
        super(Conv1d, self).__init__(
            out_channels, kernel_size, stride, padding, dilation,
            False, _single(0), groups, bias)
    
    def forward(self, x):
        in_channels = x.size(1)

        if len(self.weight) == 0:
            self._initialize_weight(in_channels)
            
        return F.conv1d(x, self.weight, self.bias, self.stride,
                      self.padding, self.dilation, self.groups)

    
class Conv2d(_ConvNd):
    '''
    Examples::

        import torch
        import torchex.nn as exnn
     
        net = exnn.Conv2d(10, 2)
        # You don't need to give the size of input for this module.
        # This network is equivalent to `nn.Conv2d(3, 10, 2)`.
     
        x = troch.randn(10, 3, 28, 28)
        y = net(x)
    '''
    def __init__(self, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1, bias=True):
        kernel_size = _pair(kernel_size)
        stride = _pair(stride)
        padding = _pair(padding)
        dilation = _pair(dilation)
        super(Conv2d, self).__init__(
            out_channels, kernel_size, stride, padding, dilation,
            False, _pair(0), groups, bias)

    def forward(self, x):
        if len(self.weight) == 0:
            in_channels = x.size(1)            
            self._initialize_weight(in_channels)            

        return F.conv2d(x, self.weight, self.bias, self.stride,
                      self.padding, self.dilation, self.groups)


class Conv3d(_ConvNd):
    '''
    Examples::

        import torch
        import torchex.nn as exnn
     
        net = exnn.Conv3d(10, 2)
        # You don't need to give the size of input for this module.
        # This network is equivalent to `nn.Conv3d(3, 10, 2)`.
     
        x = troch.randn(10, 3, 100, 28, 28)
        y = net(x)
    '''
    def __init__(self, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1, bias=True):
        kernel_size = _triple(kernel_size)
        stride = _triple(stride)
        padding = _triple(padding)
        dilation = _triple(dilation)
        super(Conv3d, self).__init__(
            out_channels, kernel_size, stride, padding, dilation,
            False, _triple(0), groups, bias)
    
    def forward(self, x):
        in_channels = x.size(1)

        if len(self.weight) == 0:
            self._initialize_weight(in_channels)                        

        return F.conv3d(x, self.weight, self.bias, self.stride,
                      self.padding, self.dilation, self.groups)
    



import torch
import torch.nn as nn
import torch.nn.functional as F

class InceptionBN(nn.Module):

    """Inception module of the new GoogLeNet with BatchNormalization.
    This chain acts like :class:`Inception`, while InceptionBN uses the
    :class:`BatchNormalization` on top of each convolution, the 5x5 convolution
    path is replaced by two consecutive 3x3 convolution applications, and the
    pooling method is configurable.
    See: `Batch Normalization: Accelerating Deep Network Training by Reducing \
    Internal Covariate Shift <https://arxiv.org/abs/1502.03167>`_.
    Args:
        in_channels (int or None): Number of channels of input arrays.
        out1 (int): Output size of the 1x1 convolution path.
        proj3 (int): Projection size of the single 3x3 convolution path.
        out3 (int): Output size of the single 3x3 convolution path.
        proj33 (int): Projection size of the double 3x3 convolutions path.
        out33 (int): Output size of the double 3x3 convolutions path.
        pooltype (str): Pooling type. It must be either ``'max'`` or ``'avg'``.
        proj_pool (int or None): Projection size in the pooling path. If
            ``None``, no projection is done.
        stride (int): Stride parameter of the last convolution of each path.
    .. seealso:: :class:`Inception`
    """

    def __init__(self, in_channels, out1, proj3, out3, proj33, out33,
                 pooltype='max', proj_pool=None, stride=1):
        super(InceptionBN, self).__init__()
        self.out1 = out1
        self.proj_pool = proj_pool
        self.stride = stride
        self.pooltype = pooltype
        if pooltype != 'max' and pooltype != 'avg':
            raise NotImplementedError()
        
        self.proj3 = nn.Conv2d(in_channels, proj3, 1, bias=False)
        self.conv3 = nn.Conv2d(proj3, out3, 3, padding=1, stride=stride, bias=False)
        self.proj33 = nn.Conv2d(in_channels, proj33, 1, bias=False)
        self.conv33a = nn.Conv2d(proj33, out33, 3, padding=1, bias=False)
        self.conv33b = nn.Conv2d(out33, out33, 3, padding=1, stride=stride, bias=False)
        self.proj3n = nn.BatchNorm2d(proj3)
        self.conv3n = nn.BatchNorm2d(out3)
        self.proj33n = nn.BatchNorm2d(proj33)
        self.conv33an = nn.BatchNorm2d(out33)
        self.conv33bn = nn.BatchNorm2d(out33)

        if out1 > 0:
            assert stride == 1
            assert proj_pool is not None
            self.conv1 = nn.Conv2d(in_channels, out1, 1, stride=stride, bias=False)
            self.conv1n = nn.BatchNorm2d(out1)

        if proj_pool is not None:
            self.poolp = nn.Conv2d(in_channels, proj_pool, 1, bias=False)
            self.poolpn = nn.BatchNorm2d(proj_pool)

    def forward(self, x):
        outs = []

        if self.out1 > 0:
            h1 = self.conv1(x)
            h1 = self.conv1n(h1)
            h1 = F.relu(h1)
            outs.append(h1)

        h3 = F.relu(self.proj3n(self.proj3(x)))
        h3 = F.relu(self.conv3n(self.conv3(h3)))
        outs.append(h3)

        h33 = F.relu(self.proj33n(self.proj33(x)))
        h33 = F.relu(self.conv33an(self.conv33a(h33)))
        h33 = F.relu(self.conv33bn(self.conv33b(h33)))
        outs.append(h33)

        if self.pooltype == 'max':
            p = F.max_pool2d(x, 3, stride=self.stride, padding=1,
                             cover_all=False)
        else:
            p = F.avg_pool2d(x, 3, stride=self.stride,
                             padding=1)
        if self.proj_pool is not None:
            p = F.relu(self.poolpn(self.poolp(p)))
        outs.append(p)

        y = torch.cat(outs, dim=1)
        return y



import torch.nn as nn
import torch.nn.functional as F

class GlobalAvgPool1d(nn.Module):
    def __init__(self):
        super(GlobalAvgPool1d, self).__init__()

    def forward(self, x):
        assert len(x.size()) == 3, x.size()
        B, C, L = x.size()
        return F.avg_pool1d(x, L)


class GlobalAvgPool2d(nn.Module):
    def __init__(self):
        super(GlobalAvgPool2d, self).__init__()

    def forward(self, x):
        assert len(x.size()) == 4, x.size()
        B, C, W, H = x.size()
        return F.avg_pool2d(x, (W, H))


class GlobalMaxPool1d(nn.Module):
    def __init__(self):
        super(GlobalMaxPool1d, self).__init__()

    def forward(self, x):
        assert len(x.size()) == 3, x.size()
        B, C, L = x.size()
        return F.avg_pool1d(x, L)

    
class GlobalMaxPool2d(nn.Module):
    def __init__(self):
        super(GlobalMaxPool2d, self).__init__()

    def forward(self, x):
        assert len(x.size()) == 4, x.size()
        B, C, W, H = x.size()
        return F.max_pool2d(x, (W, H))
    

class MaxAvgPool2d(nn.Module):
    '''
    :param kernel_size: the size of the window to take a max and average over
    :param stride: the size of stride to move kernel
    :param padding: implicit zero padding to be added on both sides
    :param dilation: a parameter that controls the stride of elements in the window
    :param return_indices: if True, will return the max indices along with the outputs. Useful when Unpooling later
    :param ceil_mode: when True, will use ceil instead of floor to compute the output shape    

    :type kernel_size: int or list
    :type stride: int or list
    '''

    def __init__(self, kernel_size, stride=None, padding=0,
                 dilation=1, return_indices=False, ceil_mode=False,
                 count_include_pad=True):
        super(MaxAvgPool2d, self).__init__()
        kwargs = dict(
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            return_indices=return_indices,
            ceil_mode=ceil_mode)
        self.max_pool = nn.MaxPool2d(**kwargs)
        del kwargs['dilation']
        del kwargs['return_indices']        
        kwargs['count_include_pad'] = count_include_pad
        self.avg_pool = nn.AvgPool2d(**kwargs)

    def forward(self, x):
        mx = self.max_pool(x)
        ax = self.avg_pool(x)
        return mx + ax



import torch
import torch.nn as nn

from ..functional import (crop_2d,
                         crop_3d)

class Crop2d(nn.Module):
    """
    :params torch.Tensor input: Input(B, C, W, H)
    """    
    def __init__(self,
                 crop_left: int=0, crop_right: int=0,
                 crop_top: int=0, crop_bottom: int=0):
        super(Crop2d, self).__init__()
        self.crop_left = crop_left
        self.crop_right = crop_right
        self.crop_top = crop_top
        self.crop_bottom = crop_bottom

    def forward(self, input):
        return crop_2d(input,
                       self.crop_left,
                       self.crop_right,
                       self.crop_top,
                       self.crop_bottom)
        
class Crop3d(nn.Module):
    """
    :params torch.Tensor input: Input(B, C, D, W, H)
    """    
    def __init__(self, crop_size):
        super(Crop3d, self).__init__()
        self.crop_size = crop_size

    def forward(self, input):
        return crop_3d(input, self.crop_size)
        



import torch.nn as nn

class LSTMStep(nn.Module):
    """Layer that performs a single step LSTM update.
    This layer performs a single step LSTM update. Note that it is *not*
    a full LSTM recurrent network. The LSTMStep layer is useful as a
    primitive for designing layers such as the AttnLSTMEmbedding or the
    IterRefLSTMEmbedding below.
    """
    def __init__(self,
                 output_dim,
                 input_dim,
                 init_fn=initializations.glorot_uniform,
                 inner_init_fn=initializations.orthogonal,
                 activation_fn=activations.tanh,
                 inner_activation_fn=activations.hard_sigmoid,
                 **kwargs):
        pass


class AttnLSTMEmbedding(nn.Module):
    """Implements AttnLSTM as in matching networks paper.
    The AttnLSTM embedding adjusts two sets of vectors, the "test" and
    "support" sets. The "support" consists of a set of evidence vectors.
    Think of these as the small training set for low-data machine
    learning.  The "test" consists of the queries we wish to answer with
    the small amounts ofavailable data. The AttnLSTMEmbdding allows us to
    modify the embedding of the "test" set depending on the contents of
    the "support".  The AttnLSTMEmbedding is thus a type of learnable
    metric that allows a network to modify its internal notion of
    distance.
    References:
    Matching Networks for One Shot Learning
    https://arxiv.org/pdf/1606.04080v1.pdf
    Order Matters: Sequence to sequence for sets
    https://arxiv.org/abs/1511.06391
    """

    def __init__(self, n_test, n_support, n_feat, max_depth, **kwargs):
        pass


class IterRefLSTMEmbedding(nn.Module):
    """Implements the Iterative Refinement LSTM.
    Much like AttnLSTMEmbedding, the IterRefLSTMEmbedding is another type
    of learnable metric which adjusts "test" and "support." Recall that
    "support" is the small amount of data available in a low data machine
    learning problem, and that "test" is the query. The AttnLSTMEmbedding
    only modifies the "test" based on the contents of the support.
    However, the IterRefLSTM modifies both the "support" and "test" based
    on each other. This allows the learnable metric to be more malleable
    than that from AttnLSTMEmbeding.
    """
    
    def __init__(self, n_test, n_support, n_feat, max_depth, **kwargs):
        pass



import torch
from torch.nn import Parameter
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
import math


class IndRNNCell(nn.Module):
    r"""An IndRNN cell with tanh or ReLU non-linearity.
    * https://arxiv.org/abs/1804.04849

    .. math::

        h' = \tanh(w_{ih} * x + b_{ih}  +  w_{hh} (*) h)
    With (*) being element-wise vector multiplication.
    If nonlinearity='relu', then ReLU is used in place of tanh.

    Args:
        input_size: The number of expected features in the input x
        hidden_size: The number of features in the hidden state h
        bias: If ``False``, then the layer does not use bias weights b_ih and b_hh.
            Default: ``True``
        nonlinearity: The non-linearity to use ['tanh'|'relu']. Default: 'relu'
        hidden_min_abs: Minimal absolute inital value for hidden weights. Default: 0
        hidden_max_abs: Maximal absolute inital value for hidden weights. Default: None

    Inputs: input, hidden
        - **input** (batch, input_size): tensor containing input features
        - **hidden** (batch, hidden_size): tensor containing the initial hidden
          state for each element in the batch.

    Outputs: h'
        - **h'** (batch, hidden_size): tensor containing the next hidden state
          for each element in the batch

    Attributes:
        weight_ih: the learnable input-hidden weights, of shape
            `(input_size x hidden_size)`
        weight_hh: the learnable hidden-hidden weights, of shape
            `(hidden_size)`
        bias_ih: the learnable input-hidden bias, of shape `(hidden_size)`

    Examples::

        >>> rnn = nn.IndRNNCell(10, 20)
        >>> input = Variable(torch.randn(6, 3, 10))
        >>> hx = Variable(torch.randn(3, 20))
        >>> output = []
        >>> for i in range(6):
        ...     hx = rnn(input[i], hx)
        ...     output.append(hx)
    """

    def __init__(self, input_size, hidden_size, bias=True, nonlinearity="relu",
                 hidden_min_abs=0, hidden_max_abs=None):
        super(IndRNNCell, self).__init__()
        self.hidden_max_abs = hidden_max_abs
        self.hidden_min_abs = hidden_min_abs
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.bias = bias
        self.nonlinearity = nonlinearity
        self.weight_ih = Parameter(torch.Tensor(hidden_size, input_size))
        self.weight_hh = Parameter(torch.Tensor(hidden_size))
        if bias:
            self.bias_ih = Parameter(torch.Tensor(hidden_size))
        else:
            self.register_parameter('bias_ih', None)
        self.reset_parameters()

    def reset_parameters(self):
        stdv = 1.0 / math.sqrt(self.hidden_size)
        for name, weight in self.named_parameters():
            if "bias" in name:
                weight.data.zero_()
            elif "weight_hh" in name:
                if self.hidden_max_abs:
                    stdv_ = self.hidden_max_abs
                else:
                    stdv_ = stdv
                weight.data.uniform_(-stdv_, stdv_)
            elif "weight_ih" in name:
                weight.data.normal_(0, 0.01)
            else:
                weight.data.normal_(0, 0.01)
                # weight.data.uniform_(-stdv, stdv)
        self.check_bounds()

    def check_bounds(self):
        if self.hidden_min_abs:
            abs_kernel = torch.abs(self.weight_hh.data)
            min_abs_kernel = torch.clamp(abs_kernel, min=self.hidden_min_abs)
            self.weight_hh.data.copy_(
                torch.mul(torch.sign(self.weight_hh.data), min_abs_kernel))
        if self.hidden_max_abs:
            self.weight_hh.data.copy_(
                torch.clamp(self.weight_hh.data, max=self.hidden_max_abs,
                            min=-self.hidden_max_abs))

    def forward(self, input, hx):
        if self.nonlinearity == "tanh":
            func = IndRNNTanhCell
        elif self.nonlinearity == "relu":
            func = IndRNNReLuCell
        else:
            raise RuntimeError(
                "Unknown nonlinearity: {}".format(self.nonlinearity))

        return func(input, hx, self.weight_ih, self.weight_hh, self.bias_ih)


def IndRNNTanhCell(input, hidden, w_ih, w_hh, b_ih=None):
    hy = F.tanh(F.linear(input, w_ih, b_ih) + F.mul(w_hh, hidden))
    return hy


def IndRNNReLuCell(input, hidden, w_ih, w_hh, b_ih=None):
    hy = F.relu(F.linear(input, w_ih, b_ih) + F.mul(w_hh, hidden))
    return hy


class IndRNN(nn.Module):
    r"""Applies a multi-layer IndRNN with `tanh` or `ReLU` non-linearity to an
    input sequence.


    For each element in the input sequence, each layer computes the following
    function:

    .. math::

        h_t = \tanh(w_{ih} x_t + b_{ih}  +  w_{hh} (*) h_{(t-1)})

    where :math:`h_t` is the hidden state at time `t`, and :math:`x_t` is
    the hidden state of the previous layer at time `t` or :math:`input_t`
    for the first layer. (*) is element-wise multiplication.
    If :attr:`nonlinearity`='relu', then `ReLU` is used instead of `tanh`.

    Args:
        input_size: The number of expected features in the input `x`
        hidden_size: The number of features in the hidden state `h`
        num_layers: Number of recurrent layers.
        nonlinearity: The non-linearity to use. Can be either 'tanh' or 'relu'. Default: 'tanh'
        bias: If ``False``, then the layer does not use bias weights `b_ih` and `b_hh`.
            Default: ``True``
        batch_first: If ``True``, then the input and output tensors are provided
            as `(batch, seq, feature)`

    Inputs: input, h_0
        - **input** of shape `(seq_len, batch, input_size)`: tensor containing the features
          of the input sequence. The input can also be a packed variable length
          sequence. See :func:`torch.nn.utils.rnn.pack_padded_sequence`
          or :func:`torch.nn.utils.rnn.pack_sequence`
          for details.
        - **h_0** of shape `(num_layers * num_directions, batch, hidden_size)`: tensor
          containing the initial hidden state for each element in the batch.
          Defaults to zero if not provided.

    Outputs: output, h_n
        - **output** of shape `(seq_len, batch, hidden_size * num_directions)`: tensor
          containing the output features (`h_k`) from the last layer of the RNN,
          for each `k`.  If a :class:`torch.nn.utils.rnn.PackedSequence` has
          been given as the input, the output will also be a packed sequence.
        - **h_n** (num_layers * num_directions, batch, hidden_size): tensor
          containing the hidden state for `k = seq_len`.

    Attributes:
        cells[k]: individual IndRNNCells containing the weights

    Examples::

        >>> rnn = nn.IndRNN(10, 20, 2)
        >>> input = torch.randn(5, 3, 10)
        >>> h0 = torch.randn(2, 3, 20)
        >>> output = rnn(input, h0)
    """

    def __init__(self, input_size, hidden_size, n_layer=1, batch_norm=False,
                 step_size=None, **kwargs):
        super(IndRNN, self).__init__()
        self.hidden_size = hidden_size
        if batch_norm and step_size is None:
            raise Exception("Frame wise batch size needs to know the step size")
        self.batch_norm = batch_norm
        self.step_size = step_size
        self.n_layer = n_layer

        cells = []
        for i in range(n_layer):
            if i == 0:
                cells += [IndRNNCell(input_size, hidden_size, **kwargs)]
            else:
                cells += [IndRNNCell(hidden_size, hidden_size, **kwargs)]
        self.cells = nn.ModuleList(cells)

        if batch_norm:
            bns = []
            for i in range(n_layer):
                bns += [nn.BatchNorm2d(step_size)]
            self.bns = nn.ModuleList(bns)
            
        
        h0 = torch.zeros(hidden_size)
        self.register_buffer('h0', torch.autograd.Variable(h0))


    def forward(self, x, hidden=None):                
        for i, cell in enumerate(self.cells):
            cell.check_bounds()
            hx = self.h0.unsqueeze(0).expand(x.size(0), self.hidden_size).contiguous()
            outputs = []
            for t in range(x.size(1)):
                x_t = x[:, t]
                hx = cell(x_t, hx)
                outputs += [hx]
            x = torch.stack(outputs, 1)
            if self.batch_norm:
                x = self.bns[i](x)
        return x.squeeze(2)



import math

import torch
import torch.nn as nn
from torch.nn.parameter import Parameter

from ..functional.local import conv2d_local
from .util import _pair

def _conv_output_length(input_length, filter_size, stride):
    output_length = input_length - filter_size + 1
    return output_length

class Conv2dLocal(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, in_size=None,
                 padding=0, bias=True):
        super(Conv2dLocal, self).__init__()        
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = _pair(kernel_size)
        self.stride = _pair(stride)
        self.in_size = in_size
        self.padding = padding
        self.bias_flag = bias
        self.bias = None
        self.weight = None

        kh, kw = _pair(self.kernel_size)
        ih, iw = _pair(in_size)

        if (self.in_channels is not None) and (self.in_size is not None):
            self._initialize_params(in_channels, in_size)
            self.reset_parameters()

    def _initialize_params(self, in_channels, in_size):
        kh, kw = _pair(self.kernel_size)
        ih, iw = _pair(in_size)
        oh = _conv_output_length(ih, kh, self.stride[0])
        ow = _conv_output_length(iw, kw, self.stride[1])
        W_shape = (self.out_channels, oh, ow, in_channels, kh, kw)
        bias_shape = (self.out_channels, oh, ow,)
        self.weight = Parameter(torch.Tensor(*W_shape))
        if self.bias_flag:
            self.bias = Parameter(torch.Tensor(*bias_shape))            
        else:
            self.register_parameter('bias', None)
        
    def reset_parameters(self):
        n = self.in_channels
        for k in self.kernel_size:
            n *= k
        stdv = 1. / math.sqrt(n)
        self.weight.data.uniform_(-stdv, stdv)
        if self.bias is not None:
            self.bias.data.uniform_(-stdv, stdv)        

    def forward(self, input):
        if self.weight is None:
            self._initialize_params(input.shape[1], input.shape[2:])
            self.reset_parameters()
        return conv2d_local(input, self.weight, self.bias, self.stride)




from .pooling import (GlobalAvgPool1d,
                      GlobalAvgPool2d,
                      GlobalMaxPool1d,
                      GlobalMaxPool2d,
                      MaxAvgPool2d)

from .cropping import (Crop2d,
                       Crop3d)

from .local import Conv2dLocal

from .util import Flatten

from .conv import (MLPConv2d,
                   UpsampleConvLayer)

from .cordconv import CordConv2d


from .dft import (DFT1d,
                  DFT2d)

from .padding import (PeriodicPad2d,
                      PeriodicPad3d)

from .highway import Highway

from .cnn_blocks import (Inception,
                         InceptionBN)

from .rnn.basic import (SequenceLinear,
                        LSTM)

from .rnn.indrnn import (IndRNNCell,
                         IndRNNTanhCell,
                         IndRNNReLuCell,
                         IndRNN)

from .graph.linear import GraphLinear
from .graph.conv import GraphConv
from .graph.conv import SparseMM
from .graph.batchnorm import GraphBatchNorm

from .lazy import *



import torch
import torch.nn as nn
import torch.nn.functional as F


class GRUDCell(nn.GRUCell):
    def __init__(self, input_size, hidden_size, bias=True):
        super(GRUDCell, self).__init__(input_size, hidden_size)
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.bias = bias
        
        self.reset_layer_i = nn.Linear(input_size, hidden_size, bias)
        self.reset_layer_h = nn.Linear(hidden_size, hidden_size, bias)
        
        self.current_memory_i = nn.Linear(input_size, hidden_size, bias)
        self.current_memory_h = nn.Linear(hidden_size, hidden_size, bias)

    def forward(self, input, hx=None):
        #self.check_forward_input(input)
        #self.check_forward_hidden(input, hx)
        z = F.sigmoid(input + hx)
        rh = F.sigmoid(self.reset_layer_i(input) + self.reset_layer_h(hx))

        cx = self.current_memory_i(input)
        ch = self.current_memory_h(hx)
        c = torch.tanh(cx + torch.mul(rh, ch))
        
        h = torch.mul(z, hx) + torch.mul(1 - z, c)
        
        return h, h

class GRUD(nn.Module):
    '''

    refs: https://www.nature.com/articles/s41598-018-24271-9
    '''
    def __init__(self, input_size, hidden_size,
                 num_layers=1, bias=True, batch_first=True,
                 dropout=0, bidirectional=False):
        # not support multi layers
        # not support bidirectional
        # not suport dropout
        super(GRUD, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.batch_first = batch_first
        self.cell = GRUDCell(input_size, hidden_size, bias)

    def forward(self, input, hx=None):
        is_packed = isinstance(input, PackedSequence)
        if is_packed:
            input, batch_sizes = input
            max_batch_size = int(batch_sizes[0])
        else:
            batch_sizes = None
            max_batch_size = input.size(0) if self.batch_first else input.size(1)

        if hx is None:
            hx = torch.autograd.Variable(input.data.new(max_batch_size,
                                                        self.hidden_size).zero_(),
                                         requires_grad=False)

        layer_output_list = []
        last_state_list   = []
            
        if self.batch_first:
            B, T, F = input.size()
        else:
            T, B, F = input.size()
        x = input
        h = hx
        for t in range(T):
            o, h = self.cell(x[:, t, :], h)
            layer_output = torch.stack(o, dim=1)
            layer_output_list.append(layer_output)

        return torch.stack(layer_output), h
    

if __name__ == "__main__":
    ## WIP
    from torch.nn.utils.rnn import PackedSequence
    
    grud = GRUD(20, 20)
    a = torch.randn(1, 10, 20)
    print(grud)
    grud(torch.autograd.Variable(a))



import math
import torch
import torch.nn as nn
import torch.nn.functional as F

from .base import LazyBase

class Linear(LazyBase):
    '''
    Examples::

        import torch
        import torchex.nn as exnn
     
        net = exnn.Linear(10)
        # You don't need to give the size of input for this module.
        # This network is equivalent to `nn.Linear(100, 10)`.
     
        x = troch.randn(10, 100)
        y = net(x)

    '''
    def __init__(self, out_features, use_bias=True):
        super(Linear, self).__init__()
        self.in_features = None
        self.out_features = out_features
        self.use_bias = use_bias
        
        self.initialize = True        
        self.weight = nn.Parameter(None)
        self.bias = nn.Parameter(None)
        self._register_load_state_dict_pre_hook(self._lazy_load_state_dict_hook)

    def _lazy_load_state_dict_hook(self, state_dict, prefix, local_metadata, strict,
                                   missing_keys, unexpected_keys, error_msgs):
        for name, data in state_dict.items():
            if prefix in name:
                if 'weight' in name:
                    self.in_features = data.shape[-1]
                    self.weight.data = data
                elif 'bias' in name:
                    self.bias.data = data
                else:
                    raise ValueError(name)

    def forward(self, x):
        if len(self.weight.data) == 0:
            self.in_features = x.shape[-1]
            self.weight.data =  torch.Tensor(self.out_features, self.in_features)
            stdv = 1. / math.sqrt(self.weight.size(1))
            self.weight.data.uniform_(-stdv, stdv)
            
            if self.use_bias:
                self.bias.data = torch.Tensor(self.out_features)
                self.bias.data.uniform_(-stdv, stdv)

            self.weight = self._to_device(self.weight)
            self.bias = self._to_device(self.bias)            
            
        return F.linear(x, self.weight, self.bias)
    
    def extra_repr(self):
        return 'in_features={}, out_features={}, bias={}'.format(
            self.in_features, self.out_features, self.bias is not None
        )    
    
if __name__ == "__main__":
    x = torch.randn(10, 2).to('cuda')
    net = Linear(10)
    net = net.to('cuda')
    y = net(x)
    print(y)



import torch

class GraphBatchNorm(torch.nn.Module):
    """Graph Batch Normalization layer.

    .. seealso:: :class:`torch.autograd.BatchNorm1d`
    """
    def __init__(self, num_features):
        super(GraphBatchNorm, self).__init__()
        self.bn = torch.nn.BatchNorm1d(num_features)

    def __call__(self, x):
        """Forward propagation.

        Args:
            x (:class:`torch.autograd.Variable`, or :class:`numpy.ndarray`
                Input array that should be a float array whose ``ndim`` is 3.

                It represents a minibatch of atoms, each of which consists
                of a sequence of molecules. Each molecule is represented
                by integer IDs. The first axis is an index of atoms
                (i.e. minibatch dimension) and the second one an index
                of molecules.

        Returns:
            :class:`touch.autograd.Variable`:
                A 3-dimeisional array.

        """
        # (minibatch, atom, ch)

        # The implemenataion of batch normalization for graph convolution below
        # is rather naive. To be precise, it is necessary to consider the
        # difference in the number of atoms for each graph. However, the
        # implementation below does not take it into account, and assumes
        # that all graphs have the same number of atoms, hence extra numbers
        # of zero are included when average is computed. In other word, the
        # results of batch normalization below is biased.

        s0, s1, s2 = x.size()
        x = x.view(s0 * s1, s2)
        x = self.bn(x)
        x = x.view(s0, s1, s2)

        return x



import collections.abc
from itertools import repeat

import torch.nn as nn

def _ntuple(n):
    def parse(x):
        if isinstance(x, collections.abc.Iterable):
            return x
        return tuple(repeat(x, n))
    return parse

_single = _ntuple(1)
_pair = _ntuple(2)
_triple = _ntuple(3)
_quadruple = _ntuple(4)


class Flatten(nn.Module):
    def forward(self, input):
        return input.view(input.size(0), -1)

class Throw(nn.Module):
    ''' Nothint to do.
    '''
    def forward(self, input):
       return input 



from typing import Tuple, Union

import torch.nn as nn
from torch import Tensor
from torch.nn.utils.rnn import PackedSequence

from ...init import rnn_init


class SequenceLinear(nn.Module):
    '''
    x[T, B, F] -> nn.Linear(x[T, BxF]) -> x[T, B, F]
    if batch_first is True,
    x[B, T, F] -> x[T, B, F] -> nn.Linear(x[T, BxF]) -> x[T, B, F]
    '''
    def __init__(self,
                 in_features,
                 out_features,
                 batch_first=True,
                 nonlinearity='sigmoid',
                 init_mean=0,
                 init_std=1,
                 init_xavier: bool=True,
                 init_normal: bool=True,
                 init_gain = None,
                 dropout=0.0,
                 bias=True,
    ):
        super(SequenceLinear, self).__init__()
        self.linear = nn.Linear(in_features, out_features, bias)
        self.batch_first = batch_first
        self.out_features = out_features
        self.nonlinearity = nonlinearity
        if not init_gain:
            init_gain = nn.init.calculate_gain(nonlinearity)
        
        feedforward_init(self.linear,
                         init_mean,
                         init_std,
                         init_xavier,
                         init_normal,
                         init_gain)
        
    def forward(self, x: Tensor):
        assert len(x.size()) == 3, x.size()
        if self.batch_first:
            B, T, F = x.size()
            x = torch.transpose(x, 0, 1)
        else:
            T, B, F = x.size()
        x = x.contiguous()
        x = x.view(T*B, F)
        x = self.linear(x)
        x = getattr(TF, self.nonlinearity)(x)
        
        if self.batch_first:
            x = x.view(T, B, self.out_features)            
            x = torch.transpose(x, 0, 1)
        else:
            x = x.view(T, B, self.out_features)
        return x



class LSTM(nn.Module):
    """Basic LSTM model. """
    def __init__(self,
                 input_size,
                 hidden_size,
                 num_layers=1,
                 bidirectional=False,
                 dropout=0.0,
                 batch_first=True,
                 init_xavier: bool=True,
                 init_normal: bool=True,
                 init_gain: float=1.0,
                 concat: bool=True,
                 init_mean=0,                 
                 init_std=0.1,
                 init_lower=0,                 
                 init_upper=0.04,
                 concat_between_layers=False,
                 ):
        super(LSTMModel, self).__init__()
        self.in_size = input_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.bidirectional = bidirectional
        self.concat = concat
        self.concat_between_layers = concat_between_layers
        
        if concat_between_layers and bidirectional:
            _rnn = []
            _rnn.append(nn.LSTM(input_size, hidden_size, num_layers=1,
                                bidirectional=bidirectional,
                                dropout=dropout, batch_first=batch_first))
            for i in range(num_layers - 1):
                _rnn.append(nn.LSTM(2 * hidden_size, hidden_size, num_layers=1,
                                    bidirectional=bidirectional,
                                    dropout=dropout, batch_first=batch_first))
            for irnn in _rnn:
                rnn_init(irnn,
                         init_xavier=init_xavier,
                         init_normal=init_normal,
                         init_gain=init_gain,
                         init_mean=init_mean,
                         init_std=init_std)
                
            self.rnn = nn.ModuleList(_rnn)
        else:
            self.rnn =\
                       nn.LSTM(input_size, hidden_size, num_layers,
                               bidirectional=bidirectional,
                               dropout=dropout, batch_first=batch_first,
                       )
            rnn_init(self.rnn,
                     init_xavier=init_xavier,
                     init_normal=init_normal,
                     init_gain=init_gain,
                     init_mean=init_mean,
                     init_std=init_std)
            

    def forward(self,
                x: Union[Tensor, PackedSequence],
                hx: Union[Tensor, Tuple[Tensor, ...]]=None) ->\
                Tuple[Tensor, Tensor]:
        
        assert isinstance(x, Tensor) or\
            isinstance(x, PackedSequence), type(x)

        if self.concat_between_layers and self.bidirectional:
            output = x
            for idx, _rnn in enumerate(self.rnn):
                _rnn.flatten_parameters()                
                output, hx = _rnn(output)
                _rnn.flatten_parameters()                
        else:
            self.rnn.flatten_parameters()
            output, hx = self.rnn(x, hx)
            self.rnn.flatten_parameters()            
        
        if (not self.concat) and self.bidirectional:
            B, T, F = output.size()
            output = output[:, :, :F//2] + output[:, :, F//2:]
            
        return output, hx



import torch

from .util import tensordot

def _pair(x):
    if hasattr(x, '__getitem__'):
        return x
    return x, x

class Conv2dLocalFunction(torch.autograd.Function):
    ''' Conv2dLocalFunction is based on chainer implementation.

    Ref: https://github.com/chainer/chainer/blob/master/chainer/functions/connection/local_convolution_2d.py
    '''
        
    @staticmethod
    def forward(ctx, input, sy, sx):
        ctx.save_for_backward(input)
        ctx.sy = sy
        ctx.sx = sx
        
        x, W = input[:2]
        b = input[2] if len(input) == 3 else None

        stride_row, stride_col = sy, sx
        output_row, output_col = W.shape[1], W.shape[2]
        feature_dim = W.shape[3] * W.shape[4] * W.shape[5]
        output = torch.empty((x.shape[0], W.shape[0], output_row, output_col,), dtype=x.dtype)
        for i in range(output_row):
            for j in range(output_col):
                slice_row = slice(i * stride_row,
                                  i * stride_row + W.shape[4])
                slice_col = slice(j * stride_col,
                                  j * stride_col + W.shape[5])
                x_flatten = torch.reshape(x[..., slice_row, slice_col],
                                       (-1, feature_dim))
                W_flatten = torch.reshape(W[:, i, j, ...],
                                       (-1, feature_dim))
                output[..., i, j] = torch.matmul(x_flatten, torch.t(W_flatten))

        if b is not None:
            output += b[None, :, :, :]

        return output


    @staticmethod
    def backward(ctx, grad_output):
        input, = ctx.saved_tensors
        x_tensor, W_tensor  = input[:2]
        x, W = x_tensor.data, W_tensor.data
        stride_row, stride_col = ctx.sy, ctx.sx
        output_row, output_col = W.shape[1], W.shape[2]
        gy_tensor = grad_output.clone()
        gy = gy_tensor.data
        ret = []

        gx = torch.zero_like(x)
        for i in range(output_row):
            for j in range(output_col):
                slice_row = slice(i * stride_row,
                                  i * stride_row + W.shape[4])
                slice_col = slice(j * stride_col,
                                  j * stride_col + W.shape[5])
                # ochans * ichans * krows * kcols
                W_slice = W[:, i, j, ...]
                # nsamps * ochans
                gy_slice = gy[..., i, j]
                # -> nsamps * ichans * krows * kcols
                gx[:, :, slice_row, slice_col] += tensordot(
                    gy_slice, W_slice, axes=[(1,), (0,)]
                )
                # use torch.mm
            ret.append(gx.type(x.dtype))            
        return ret


def conv2d_local(x, W, b=None, stride=1):
    fnode = Conv2dLocalFunction(stride)
    sx, sy = _pair(stride)
    if b is None:
        args = (x, W)
    else:
        args = (x, W, b)
    y = fnode.apply(args, sy, sx)
    return y    


if __name__ == '__main__':
    x = torch.randn((2, 3, 7, 7), requires_grad=True)    
    W = torch.randn((2, 5, 5, 3, 3, 3), requires_grad=True)    
    b = torch.randn((2, 5, 5), requires_grad=True)    

    y = conv2d_local(x, W, b)
    t = torch.randn((2, 2, 5, 5), requires_grad=True)
    loss = y - t
    loss = torch.sum(loss)
    loss.backward()
    y.backward()
    from torch.autograd import gradcheck
    print(gradcheck(conv2d_local, [x, W]))
    



import torch
import torch.nn as nn

# TODO: implement 1d and 3d.

class CordConv2d(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1, bias=True, with_r=False):
        super(CordConv2d, self).__init__()
        self.with_r = with_r

        if self.with_r:
            in_channels += 3
        else:
            in_channels += 2
            
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size, stride,
                              padding, dilation, groups, bias)
        
    def forward(self, x):
        B, C, H, W = x.shape
        _c_H = torch.arange(0., H, dtype=x.dtype)
        _c_W = torch.arange(0., W, dtype=x.dtype)
        _c_H = torch.t(_c_H.repeat(W, B)).view(B, 1, H, W) / (H - 1)
        _c_W = torch.t(_c_W.repeat(H, B).t()).view(B, 1, H, W) / (W - 1)
        _c_H = _c_H * 2 - 1
        _c_W = _c_W * 2 - 1

        x = torch.cat([x, _c_H, _c_W], dim=1)
        
        if self.with_r:
           rr = torch.sqrt(torch.mul(_c_H-0.5, _c_H-0.5) + torch.mul(_c_W-0.5, _c_W-0.5))
           x = torch.cat([x, rr], dim=1)
        
        return self.conv(x)
        




def crop_2d(input,
            crop_left: int=0, crop_right: int=0,
            crop_top: int=0, crop_bottom: int=0):
    assert input.dim() == 4, 'only support Input(B, C, W, H)'
    B, C, W, H = input.size()
    return input[:, :,
                 crop_left:(W-crop_right),
                 crop_bottom:(H-crop_top)]

def crop_3d(input, crop_size: int=0):
    assert input.dim() == 5, 'only support Input(B, C, D, W, H)'
    B, C, D, W, H = input.size()
    return input[:, :,
                 crop_size:(D-crop_size),
                 crop_size:(W-crop_size),
                 crop_size:(H-crop_size)]

    



import torch.nn as nn


class GraphSortPool(nn.Module):
    '''

    Refs: https://www.cse.wustl.edu/~ychen/public/DGCNN.pdf    
    '''
    def __init__(self):
        super(GraphSortPool, self).__init__()
        pass
        


class GraphMaxPool(nn.Module):
    ''' GraphMaxPool wraps MaxPool1d.
    '''
    def __init__(self, kernel_size, stride=None, padding=0, dilation=1, return_indices=False, ceil_mode=False):
        super(GraphPool, self).__init__()
        self.pool = nn.MaxPool1d(kernel_size,
                                 stride,
                                 padding,
                                 dilation,
                                 return_indices,
                                 ceil_mode)

    def forward(self, x):
        return self.pool(x)


class GraphAvgPool(nn.Module):
    ''' GraphAvgPool wraps AvgPool1d.
    '''
    def __init__(self, kernel_size, stride=None, padding=0, dilation=1, return_indices=False, ceil_mode=False):
        super(GraphPool, self).__init__()
        self.pool = nn.AvgPool1d(kernel_size,
                                 stride,
                                 padding,
                                 dilation,
                                 return_indices,
                                 ceil_mode)

    def forward(self, x):
        return self.pool(x)
    



import torch


def save(model: torch.nn.Module, path: str) -> None:
    torch.save(model.state_dict(), path)


def load(model: torch.nn.Module, path: str) -> torch.nn.Module:
    model.load_state_dict(torch.load(path))
    return model



import torch
import torch.nn as nn

class PeriodicPad2d(nn.Module):
    """

    :params torch.Tensor input: Input(B, C, W, H)

    # https://github.com/ZichaoLong/aTEAM/blob/master/nn/functional/utils.py
    """
    
    def __init__(self,
                 pad_left: int=0, pad_right: int=0,
                 pad_top: int=0, pad_bottom: int=0):
        super(PeriodicPad2d, self).__init__()
        self.__doc__ = 'hello'
        self.pad_left = pad_left
        self.pad_right = pad_right
        self.pad_top = pad_top
        self.pad_bottom = pad_bottom

    def forward(self, input):
        assert input.dim() == 4, 'only supports Input(B, C, H, W)'
        
        B, C, H, W = input.size()
        
        left_pad = input[:, :, :, W-(self.pad_left):]
        right_pad = input[:, :, :, :self.pad_right]
        input = torch.cat([left_pad, input, right_pad], dim=3)
        top_pad = input[:, :, H-(self.pad_top):, :]
        bottom_pad = input[:, :, :self.pad_bottom, :]
        input = torch.cat([top_pad, input, bottom_pad], dim=2)
        return input


class PeriodicPad3d(nn.Module):
    '''
    Only support isotropic padding
    '''
    def __init__(self, pad: int=0):
        super(PeriodicPad3d, self).__init__()
        self.pad = pad
        
    def forward(self, input):
        '''
        :params torch.Tensor input: Input(B, C, D, W, H)    
        '''
        assert input.dim() == 5, 'only support Input(B, C, D, W, H)'
        
        B, C, D, H, W = input.size()        
        pad_0 = input[:, :, D-(self.pad):, :, :]
        pad_1 = input[:, :, :self.pad, :, :]
        input = torch.cat([pad_0, input, pad_1], dim=2)
        pad_0 = input[:, :, :, H-(self.pad):, :]
        pad_1 = input[:, :, :, :self.pad, :]
        input = torch.cat([pad_0, input, pad_1], dim=3)
        pad_0 = input[:, :, :, :, W-(self.pad):]
        pad_1 = input[:, :, :, :, :self.pad]
        input = torch.cat([pad_0, input, pad_1], dim=4)
        return input

if __name__ == '__main__':

    x = torch.range(1, 25).view(1, 1, 5, 5)
    print(x)    
    pad = PeriodicPad2d(2, 2, 2, 1)
    print(pad(x))
    print(pad(x).shape)    
    
    x = torch.range(1, 27).view(1, 1, 3, 3, 3)
    pad = PeriodicPad3d(1)
    print(pad(x))



def tensordot(a, b, axes=2):
    # code adapted from numpy
    # https://gist.github.com/deanmark/9aec75b7dc9fa71c93c4bc85c5438777
    try:
        iter(axes)
    except Exception:
        axes_a = list(range(-axes, 0))
        axes_b = list(range(0, axes))
    else:
        axes_a, axes_b = axes
    try:
        na = len(axes_a)
        axes_a = list(axes_a)
    except TypeError:
        axes_a = [axes_a]
        na = 1
    try:
        nb = len(axes_b)
        axes_b = list(axes_b)
    except TypeError:
        axes_b = [axes_b]
        nb = 1
    
    # uncomment in pytorch >= 0.5
    # a, b = torch.as_tensor(a), torch.as_tensor(b)
    as_ = a.shape
    nda = a.dim()
    bs = b.shape
    ndb = b.dim()
    equal = True
    if na != nb:
        equal = False
    else:
        for k in range(na):
            if as_[axes_a[k]] != bs[axes_b[k]]:
                equal = False
                break
            if axes_a[k] < 0:
                axes_a[k] += nda
            if axes_b[k] < 0:
                axes_b[k] += ndb
    if not equal:
        raise ValueError("shape-mismatch for sum")

    # Move the axes to sum over to the end of "a"
    # and to the front of "b"
    notin = [k for k in range(nda) if k not in axes_a]
    newaxes_a = notin + axes_a
    N2 = 1
    for axis in axes_a:
        N2 *= as_[axis]
    newshape_a = (int(np.multiply.reduce([as_[ax] for ax in notin])), N2)
    olda = [as_[axis] for axis in notin]

    notin = [k for k in range(ndb) if k not in axes_b]
    newaxes_b = axes_b + notin
    N2 = 1
    for axis in axes_b:
        N2 *= bs[axis]
    newshape_b = (N2, int(np.multiply.reduce([bs[ax] for ax in notin])))
    oldb = [bs[axis] for axis in notin]

    at = a.permute(newaxes_a).reshape(newshape_a)
    bt = b.permute(newaxes_b).reshape(newshape_b)

    res = at.matmul(bt)
    return res.reshape(olda + oldb)



import torch

def to_complex(x: torch.Tensor) -> torch.Tensor:
    '''
    change tensor from real to complex.
    note: to_complex(to_complex(x)) is not supported.
    '''
    _x = torch.zeros_like(x)
    return torch.stack([x, _x], dim=x.dim())

def to_real(x: torch.Tensor) -> torch.Tensor:
    '''
    change tensor from complex to real.    
    '''
    assert 2 == x.size(-1), 'last dimmension of complex tensor must be2 (x.size(-1) == 2)'
    return x[..., 0]

def complex_norm(x: torch.Tensor) -> torch.Tensor:
    x = x.type(torch.float)
    assert 2 == x.size(-1)
    return torch.sqrt(torch.pow(x[..., 0], 2) + torch.pow(x[..., 1], 2)).type(x.dtype)



import torch.nn as nn

class Classifier(nn.Module):
    '''
    When you use multi-GPU, This module provides calculating loss function on each GPUs.
    If you don't use this module, I know the training speed slightly decrease.
    '''
    def __init__(self, model: nn.Module, criterion: nn.Module):
        super(Classifier, self).__init__()
        self.model = model
        self.criterion = criterion

    def __call__(self, x, t):
        out = self.model(x)
        loss = self.criterion(out, t)
        return loss



import torch
import torch.nn as nn


class LazyBase(nn.Module):
    def __init__(self):
        super(LazyBase, self).__init__()
        self.to_args = None
        self.to_kwargs = None
        self._fn = []

    def to(self, *args, **kwargs):
        self.to_args = args
        self.to_kwargs = kwargs
        return super().to(*args, **kwargs)
        
    def _apply(self, fn):
        self._fn.append(fn)
        super()._apply(fn)
        return self
    
    def _to_device(self, param):
        if param is None:
            return None
        
        if self.to_args is not None:
            param.data = param.data.to(*self.to_args, **self.to_kwargs)

        if self._fn:
            for f in self._fn:
                if param is not None:
                    # Tensors stored in modules are graph leaves, and we don't
                    # want to create copy nodes, so we have to unpack the data.
                    param.data = f(param.data)
                    if param._grad is not None:
                        param._grad.data = f(param._grad.data)            
        return param



import torch
import torch.nn as nn
import numpy as np


class IntegratedGradients(object):
    '''

    Reference: https://arxiv.org/abs/1703.01365
    '''
    def __init__(self,
                 model: nn.Module,
                 divided_number: int=100):
        self.model = model
        self.divided_number = divided_number

    def __call__(self, x, label_idx: int, x_base=None):
        self.model.train()
        ig = torch.zeros_like(x)
        
        if x_base is None:
            x_base = torch.zeros_like(x)
            
        if not x.requires_grad:
            x.requires_grad = True
            
        for k in range(self.divided_number):
            self.model.zero_grad()
            x.grad = torch.zeros_like(x)
            alpha = (k + 1) / float(self.divided_number)
            y = self.model(x_base + (x - x_base) * alpha)
            y[:, label_idx].backward()
            ig += (x - x_base) * x.grad / self.divided_number
        self.ig = ig
        return ig

    def check(self):
        return torch.sum(self.ig)
    
    @classmethod
    def heatmap(cls, x:torch.Tensor, filename):
        import matplotlib.pyplot as plt
        from mpl_toolkits.axes_grid1 import make_axes_locatable
        
        x = x.data.numpy()
        
        plt.switch_backend('Agg')
        fig = plt.figure()
        ax = fig.add_subplot(111)
        absmax = np.max(np.abs(x))
        im = ax.imshow(x, aspect='equal', cmap=plt.get_cmap('bwr'),
                       vmin=-absmax, vmax=absmax)
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        plt.colorbar(im, cax=cax)
        plt.savefig(filename)
        
        
        



import math
import numbers
import warnings

import torch
import torch.nn as nn
from torch.nn.utils.rnn import PackedSequence

from .base import LazyBase


class LazyRNNBase(LazyBase):
    def __init__(self, mode, hidden_size,
                 num_layers=1, bias=True, batch_first=False,
                 dropout=0, bidirectional=False):
        super(LazyRNNBase, self).__init__()
        self.mode = mode
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.bias = bias
        self.batch_first = batch_first
        self.dropout = dropout
        self.dropout_state = {}
        self.bidirectional = bidirectional
        self.num_directions = 2 if bidirectional else 1

        if not isinstance(dropout, numbers.Number) or not 0 <= dropout <= 1 or \
                isinstance(dropout, bool):
            raise ValueError("dropout should be a number in range [0, 1] "
                             "representing the probability of an element being "
                             "zeroed")
        if dropout > 0 and num_layers == 1:
            warnings.warn("dropout option adds dropout after all but last "
                          "recurrent layer, so non-zero dropout expects "
                          "num_layers greater than 1, but got dropout={} and "
                          "num_layers={}".format(dropout, num_layers))

        if mode == 'LSTM':
            self.gate_size = 4 * hidden_size
        elif mode == 'GRU':
            self.gate_size = 3 * hidden_size
        else:
            self.gate_size = hidden_size
            
        self._all_weights = []
        self._data_ptrs = None

    def flatten_parameters(self):
        """Resets parameter data pointer so that they can use faster code paths.
        Right now, this works only if the module is on the GPU and cuDNN is enabled.
        Otherwise, it's a no-op.
        """
        any_param = next(self.parameters()).data
        if not any_param.is_cuda or not torch.backends.cudnn.is_acceptable(any_param):
            self._data_ptrs = []
            return

        # If any parameters alias, we fall back to the slower, copying code path. This is
        # a sufficient check, because overlapping parameter buffers that don't completely
        # alias would break the assumptions of the uniqueness check in
        # Module.named_parameters().
        unique_data_ptrs = set(p.data_ptr() for l in self.all_weights for p in l)
        if len(unique_data_ptrs) != sum(len(l) for l in self.all_weights):
            self._data_ptrs = []
            return

        with torch.cuda.device_of(any_param):
            import torch.backends.cudnn.rnn as rnn

            weight_arr = list(itertools.chain.from_iterable(self.all_weights))
            weight_stride0 = len(self.all_weights[0])

            # NB: This is a temporary hack while we still don't have Tensor
            # bindings for ATen functions
            with torch.no_grad():
                # NB: this is an INPLACE function on weight_arr, that's why the
                # no_grad() is necessary.
                weight_buf = torch._cudnn_rnn_flatten_weight(
                    weight_arr, weight_stride0,
                    self.input_size, rnn.get_cudnn_mode(self.mode), self.hidden_size, self.num_layers,
                    self.batch_first, bool(self.bidirectional))

            self._param_buf_size = weight_buf.size(0)
            self._data_ptrs = list(p.data.data_ptr() for p in self.parameters())

    def _apply(self, fn):
        ret = super(RNNBase, self)._apply(fn)
        self.flatten_parameters()
        return ret

    def reset_parameters(self):
        stdv = 1.0 / math.sqrt(self.hidden_size)
        for name, p in self.named_parameters():
            if 'bias' in name:
                p.data.fill_(0)
                if self.mode == "LSTM":
                    n = p.nelement()
                    p.data[n // 4:n // 2].fill_(1)  # forget bias
            else:
                p.data.uniform_(-stdv, stdv)            
            
    def _initialize_weight(self, input_size):
        for layer in range(self.num_layers):
            for direction in range(self.num_directions):
                layer_input_size = input_size if layer == 0 else self.hidden_size * self.num_directions

                w_ih = torch.Tensor(self.gate_size, layer_input_size)
                w_hh = torch.Tensor(self.gate_size, self.hidden_size)
                b_ih = torch.Tensor(self.gate_size)
                b_hh = torch.Tensor(self.gate_size)
                if self.to_args:
                    w_ih = w_ih.to(*self.to_args, **self.to_kwargs)
                    w_hh = w_hh.to(*self.to_args, **self.to_kwargs)
                    b_ih = b_ih.to(*self.to_args, **self.to_kwargs)
                    b_hh = b_hh.to(*self.to_args, **self.to_kwargs)

                w_ih = nn.Parameter(w_ih)
                w_hh = nn.Parameter(w_hh)
                b_ih = nn.Parameter(b_ih)
                b_hh = nn.Parameter(b_hh)
                
                layer_params = (w_ih, w_hh, b_ih, b_hh)

                suffix = '_reverse' if direction == 1 else ''
                param_names = ['weight_ih_l{}{}', 'weight_hh_l{}{}']
                if self.bias:
                    param_names += ['bias_ih_l{}{}', 'bias_hh_l{}{}']
                param_names = [x.format(layer, suffix) for x in param_names]

                for name, param in zip(param_names, layer_params):
                    setattr(self, name, param)
                self._all_weights.append(param_names)

        self.flatten_parameters()
        self.reset_parameters()        

    def check_forward_args(self, input, hidden, batch_sizes):
        is_input_packed = batch_sizes is not None
        expected_input_dim = 2 if is_input_packed else 3
        if input.dim() != expected_input_dim:
            raise RuntimeError(
                'input must have {} dimensions, got {}'.format(
                    expected_input_dim, input.dim()))
        if self.input_size != input.size(-1):
            raise RuntimeError(
                'input.size(-1) must be equal to input_size. Expected {}, got {}'.format(
                    self.input_size, input.size(-1)))

        if is_input_packed:
            mini_batch = int(batch_sizes[0])
        else:
            mini_batch = input.size(0) if self.batch_first else input.size(1)

        num_directions = 2 if self.bidirectional else 1
        expected_hidden_size = (self.num_layers * num_directions,
                                mini_batch, self.hidden_size)

        def check_hidden_size(hx, expected_hidden_size, msg='Expected hidden size {}, got {}'):
            if tuple(hx.size()) != expected_hidden_size:
                raise RuntimeError(msg.format(expected_hidden_size, tuple(hx.size())))

        if self.mode == 'LSTM':
            check_hidden_size(hidden[0], expected_hidden_size,
                              'Expected hidden[0] size {}, got {}')
            check_hidden_size(hidden[1], expected_hidden_size,
                              'Expected hidden[1] size {}, got {}')
        else:
            check_hidden_size(hidden, expected_hidden_size)

    def forward(self, input, hx=None):
        is_packed = isinstance(input, PackedSequence)

        if not self._all_weights:
            self.input_size = input.size(-1)
            self._initialize_weight(self.input_size)
            
        if is_packed:
            input, batch_sizes = input
            max_batch_size = int(batch_sizes[0])
        else:
            batch_sizes = None
            max_batch_size = input.size(0) if self.batch_first else input.size(1)

        if hx is None:
            num_directions = 2 if self.bidirectional else 1
            hx = input.new_zeros(self.num_layers * num_directions,
                                 max_batch_size, self.hidden_size,
                                 requires_grad=False)
            if self.mode == 'LSTM':
                hx = (hx, hx)

        has_flat_weights = list(p.data.data_ptr() for p in self.parameters()) == self._data_ptrs
        if has_flat_weights:
            first_data = next(self.parameters()).data
            assert first_data.storage().size() == self._param_buf_size
            flat_weight = first_data.new().set_(first_data.storage(), 0, torch.Size([self._param_buf_size]))
        else:
            flat_weight = None

        self.check_forward_args(input, hx, batch_sizes)
        func = self._backend.RNN(
            self.mode,
            self.input_size,
            self.hidden_size,
            num_layers=self.num_layers,
            batch_first=self.batch_first,
            dropout=self.dropout,
            train=self.training,
            bidirectional=self.bidirectional,
            dropout_state=self.dropout_state,
            variable_length=is_packed,
            flat_weight=flat_weight
        )
        output, hidden = func(input, self.all_weights, hx, batch_sizes)
        if is_packed:
            output = PackedSequence(output, batch_sizes)
        return output, hidden

    def extra_repr(self):
        s = '{input_size}, {hidden_size}'
        if self.num_layers != 1:
            s += ', num_layers={num_layers}'
        if self.bias is not True:
            s += ', bias={bias}'
        if self.batch_first is not False:
            s += ', batch_first={batch_first}'
        if self.dropout != 0:
            s += ', dropout={dropout}'
        if self.bidirectional is not False:
            s += ', bidirectional={bidirectional}'
        return s.format(**self.__dict__)

    def __setstate__(self, d):
        super(RNNBase, self).__setstate__(d)
        self.__dict__.setdefault('_data_ptrs', [])
        if 'all_weights' in d:
            self._all_weights = d['all_weights']
        if isinstance(self._all_weights[0][0], str):
            return
        num_layers = self.num_layers
        num_directions = 2 if self.bidirectional else 1
        self._all_weights = []
        for layer in range(num_layers):
            for direction in range(num_directions):
                suffix = '_reverse' if direction == 1 else ''
                weights = ['weight_ih_l{}{}', 'weight_hh_l{}{}', 'bias_ih_l{}{}', 'bias_hh_l{}{}']
                weights = [x.format(layer, suffix) for x in weights]
                if self.bias:
                    self._all_weights += [weights]
                else:
                    self._all_weights += [weights[:2]]

    @property
    def all_weights(self):
        return [[getattr(self, weight) for weight in weights] for weights in self._all_weights]




import torch

class GraphLinear(torch.nn.Module):
    """Graph Linear layer.

    This function assumes its input is 3-dimensional.
    Differently from :class:`chainer.functions.linear`, it applies an affine
    transformation to the third axis of input `x`.

    .. seealso:: :class:`torch.nn.Linear`
    """
    def __init__(self, *argv, **kwargs):

        super(GraphLinear, self).__init__()
        self.linear = torch.nn.Linear(*argv, **kwargs)

    def __call__(self, x):
        """Forward propagation.

        Args:
            x (:class:`torch.Tensor`)
                Input array that should be a float array whose ``dim`` is 3.

        Returns:
            :class:`torch.Tensor`:
                A 3-dimeisional array.

        """
        # (minibatch, atom, ch)
        s0, s1, s2 = x.size()
        x = x.view(s0 * s1, s2)
        x = self.linear(x)
        x = x.view(s0, s1, -1)
        return x



import torch.nn 

def get_module(module_name: str):
    """only support pytorch.nn modules
    """
    module_name = module_name.lower()
    print(torch.nn.modules.__all__)
    


if __name__ == '__main__':
    get_module('linear')



import torch
from torch.nn.parameter import Parameter

class SparseMM(torch.autograd.Function):
    """
    Sparse x dense matrix multiplication with autograd support.
    Implementation by Soumith Chintala:
    https://discuss.pytorch.org/t/
    does-pytorch-support-autograd-on-sparse-matrix/6156/7
    """

    def __init__(self, sparse):
        super(SparseMM, self).__init__()
        self.sparse = sparse

    def forward(self, dense):
        return torch.mm(self.sparse, dense)

    def backward(self, grad_output):
        grad_input = None
        if self.needs_input_grad[0]:
            grad_input = torch.mm(self.sparse.t(), grad_output)
        return grad_input


class GraphConv(torch.nn.Module):
    """
    Simple GCN layer, similar to https://arxiv.org/abs/1609.02907
    """

    def __init__(self, in_features, out_features, bias=True):
        super(GraphConv, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = Parameter(torch.Tensor(in_features, out_features))
        if bias:
            self.bias = Parameter(torch.Tensor(out_features))
        else:
            self.register_parameter('bias', None)
        
    def forward(self, input, adj):
        support = torch.mm(input, self.weight)
        output = SparseMM(adj)(support)
        if self.bias is not None:
            return output + self.bias
        else:
            return output

    def __repr__(self):
        return self.__class__.__name__ + ' (' \
            + str(self.in_features) + ' -> ' \
            + str(self.out_features) + ')'        



import os
import re
import setuptools
from pathlib import Path

p = Path(__file__)

def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)

version = get_version('torchex')

setup_requires = [
    'numpy',
    'pytest-runner'
]

install_requires = [
    'tensorboardx'
]
test_require = [
    'pytest-cov',
    'pytest-html',
    'pytest'
]

setuptools.setup(
    name="torchex",
    version=version,
    python_requires='>3.5',    
    author="Koji Ono",
    author_email="koji.ono@exwzd.com",
    description="Pytorch Extension Module.",
    url='https://github.com/0h-n0/torchex',
    long_description=(p.parent / 'README.md').open(encoding='utf-8').read(),
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=test_require,
    extras_require={
        'docs': [
            'sphinx >= 1.4',
            'sphinx_rtd_theme']},
    classifiers=[
        'Programming Language :: Python :: 3.6',
    ],
)



import torch
import torch.nn as nn
import torch.nn.functional as F


class Inception(nn.Module):
    """Inception module of GoogLeNet.
    It applies four different functions to the input array and concatenates
    their outputs along the channel dimension. Three of them are 2D
    convolutions of sizes 1x1, 3x3 and 5x5. Convolution paths of 3x3 and 5x5
    sizes have 1x1 convolutions (called projections) ahead of them. The other
    path consists of 1x1 convolution (projection) and 3x3 max pooling.
    The output array has the same spatial size as the input. In order to
    satisfy this, Inception module uses appropriate padding for each
    convolution and pooling.
    See: `Going Deeper with Convolutions <https://arxiv.org/abs/1409.4842>`_.
    Args:
        in_channels (int or None): Number of channels of input arrays.
        out1 (int): Output size of 1x1 convolution path.
        proj3 (int): Projection size of 3x3 convolution path.
        out3 (int): Output size of 3x3 convolution path.
        proj5 (int): Projection size of 5x5 convolution path.
        out5 (int): Output size of 5x5 convolution path.
        proj_pool (int): Projection size of max pooling path.

    TODO: Implement dimension reduction version.
    """

    def __init__(self, in_channels, out1, proj3, out3, proj5, out5, proj_pool):
        super(Inception, self).__init__()
        self.conv1 = nn.Conv2d(in_channels, out1, 1)
        self.proj3 = nn.Conv2d(in_channels, proj3, 1)
        self.conv3 = nn.Conv2d(proj3, out3, 3, padding=1)
        self.proj5 = nn.Conv2d(in_channels, proj5, 1)
        self.conv5 = nn.Conv2d(proj5, out5, 5, padding=2)
        self.projp = nn.Conv2d(in_channels, proj_pool, 1)

    def forward(self, x):
        """Computes the output of the Inception module.
        Args:
            x (~torch.Tensor): Input variable.
        Returns:
            Variable: Output variable. Its array has the same spatial size and
            the same minibatch size as the input array. The channel dimension
            has size ``out1 + out3 + out5 + proj_pool``.
        """
        out1 = self.conv1(x)
        out3 = self.conv3(F.relu(self.proj3(x)))
        out5 = self.conv5(F.relu(self.proj5(x)))
        pool = self.projp(F.max_pool2d(x, 3, stride=1, padding=1))
        y = F.relu(torch.cat((out1, out3, out5, pool), dim=1))
        return y



import torch
import torch.nn as nn
import torch.nn.functional as F


class Highway(nn.Module):

    """Highway module.
    In highway network, two gates are added to the ordinal non-linear
    transformation (:math:`H(x) = activate(W_h x + b_h)`).
    One gate is the transform gate :math:`T(x) = \\sigma(W_t x + b_t)`, and the
    other is the carry gate :math:`C(x)`.
    For simplicity, the author defined :math:`C = 1 - T`.
    Highway module returns :math:`y` defined as
    .. math::
        y = activate(W_h x + b_h) \\odot \\sigma(W_t x + b_t) +
        x \\odot(1 - \\sigma(W_t x + b_t))
    The output array has the same spatial size as the input. In order to
    satisfy this, :math:`W_h` and :math:`W_t` must be square matrices.
    Args:
        in_out_features (int): Dimension of input and output vectors.
        bias (bool): If ``True``, then this function does use the bias.
        activate: Activation function of plain array. :math:`tanh` is also
            available.
    See:
        `Highway Networks <https://arxiv.org/abs/1505.00387>`_.
    """

    def __init__(self, in_out_features, bias=True, activate=F.relu):
        super(Highway, self).__init__()
        self.in_out_features = in_out_features
        self.bias = bias
        self.activate = activate

        self.plain = nn.Linear(self.in_out_features, self.in_out_features, bias=bias)
        self.transform = nn.Linear(self.in_out_features, self.in_out_features, bias=bias)


    def forward(self, x):
        """Computes the output of the Highway module.
        Args:
            x (~torch.Tensor): Input variable.
        Returns:
            Variable: Output variable. Its array has the same spatial size and
            the same minibatch size as the input array.
        """
        out_plain = self.activate(self.plain(x))
        out_transform = torch.sigmoid(self.transform(x))
        y = out_plain * out_transform + x * (1 - out_transform)
        return y



from tensorboardX import SummaryWriter


class TensorBoarder(object):
    def __init__(self):
        self.writer = SummaryWriter()

    def histgram(self, module, n_iter):
        for name, param in module.named_parameters():
            self.writer.add_histogram(name, param.clone().cpu().data.numpy(), n_iter)





from .tensorboard import TensorBoarder

from .complex_helper import (to_complex,
                             to_real,
                             complex_norm)

from .io import (load,
                 save)



import torch
import torch.nn as nn

from ...utils import (to_complex,
                      to_real,
                      complex_norm)


class DFT1d(nn.Module):
    def __init__(self, in_channels, out_channels, normlized=True, absolute=True):
        super(DFT1d, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels

        self.absolute = absolute

    def forward(self, x):
        if self.absolute:
            x = torch.abs(x)
        return x

    
class DFT2d(nn.Module):
    def __init__(self, normlized=True, norm=True, concatinate=False):
        super(DFT2d, self).__init__()        
        self.normlized = normlized
        self.norm = norm
        self.concatinate = concatinate

    def forward(self, x):
        if x.size(-1) != 2:
            x = to_complex(x)
        x = torch.fft(x, 2, self.normlized)
        if self.concatinate:
            _dim = x[..., 1].dim() - 1
            x = torch.cat([x[..., 0], x[..., 1]], dim=_dim)
        if self.norm and not self.concatinate:
            x = complex_norm(x)
        return x

    
class DFT3d(nn.Module):
    def __init__(self, in_channels, out_channels):
        self.in_channels = in_channels
        self.out_channels = out_channels

    def forward(self):
        pass


class iDFT1d(nn.Module):
    def __init__(self, in_channels, out_channels):
        self.in_channels = in_channels
        self.out_channels = out_channels

    def forward(self):
        pass


class iDFT2d(nn.Module):
    def __init__(self, in_channels, out_channels):
        self.in_channels = in_channels
        self.out_channels = out_channels

    def forward(self):
        pass


class iDFT3d(nn.Module):
    def __init__(self, in_channels, out_channels):
        self.in_channels = in_channels
        self.out_channels = out_channels

    def forward(self):
        pass
    

class RFFT1d(nn.Module):
    def __init__(self, in_channels, out_channels):
        self.in_channels = in_channels
        self.out_channels = out_channels

    def forward(self):
        pass


class RFFT2d(nn.Module):
    def __init__(self, in_channels, out_channels):
        self.in_channels = in_channels
        self.out_channels = out_channels

    def forward(self):
        pass


class RFFT3d(nn.Module):
    def __init__(self, in_channels, out_channels):
        self.in_channels = in_channels
        self.out_channels = out_channels

    def forward(self):
        pass


class iRFFT1d(nn.Module):
    def __init__(self, in_channels, out_channels):
        self.in_channels = in_channels
        self.out_channels = out_channels

    def forward(self):
        pass


class iRFFT2d(nn.Module):
    def __init__(self, in_channels, out_channels):
        self.in_channels = in_channels
        self.out_channels = out_channels

    def forward(self):
        pass


class iRFFT3d(nn.Module):
    def __init__(self, in_channels, out_channels):
        self.in_channels = in_channels
        self.out_channels = out_channels

    def forward(self):
        pass

    
    
class iSTFT1d(nn.Module):
    def __init__(self, in_channels, out_channels):
        self.in_channels = in_channels
        self.out_channels = out_channels

    def forward(self):
        pass


class iSTFT2d(nn.Module):
    def __init__(self, in_channels, out_channels):
        self.in_channels = in_channels
        self.out_channels = out_channels

    def forward(self):
        pass



if __name__ == '__main__':
    x = torch.randn(10, 1, 28, 28)
    dft2d = DFTConv2d(1, 3)
    y = dft2d(x)
    print(y.shape)
